# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import os
import os.path
from unittest.mock import Mock

from ofxparse import OfxParser

from ledgerautosync.ledgerwrap import Ledger
from ledgerautosync.sync import CsvSynchronizer, OfxSynchronizer


def test_fresh_sync():
    ledger = Ledger(os.path.join("fixtures", "empty.lgr"))
    sync = OfxSynchronizer(ledger)
    with open(os.path.join("fixtures", "checking.ofx"), "rb") as ofx_file:
        ofx = OfxParser.parse(ofx_file)
        txns1 = ofx.account.statement.transactions
        txns2 = sync.filter(txns1, ofx.account.account_id)
        assert txns1 == txns2


def test_sync_order():
    ledger = Ledger(os.path.join("fixtures", "empty.lgr"))
    sync = OfxSynchronizer(ledger)
    with open(os.path.join("fixtures", "checking_order.ofx"), "rb") as ofx_file:
        ofx = OfxParser.parse(ofx_file)
        txns = sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
        assert txns[0].date < txns[1].date
        assert txns[1].date < txns[2].date


def test_fully_synced():
    ledger = Ledger(os.path.join("fixtures", "checking.lgr"))
    sync = OfxSynchronizer(ledger)
    ofx = OfxSynchronizer.parse_file(os.path.join("fixtures", "checking.ofx"))
    txns = sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert txns == []


def test_partial_sync():
    ledger = Ledger(os.path.join("fixtures", "checking-partial.lgr"))
    sync = OfxSynchronizer(ledger)
    ofx = OfxSynchronizer.parse_file(os.path.join("fixtures", "checking.ofx"))
    txns = sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert len(txns) == 1


def test_no_new_txns():
    ledger = Ledger(os.path.join("fixtures", "checking.lgr"))
    acct = Mock()
    acct.download = Mock(
        return_value=open(os.path.join("fixtures", "checking.ofx"), "rb")
    )
    sync = OfxSynchronizer(ledger)
    assert len(sync.get_new_txns(acct, 7, 7)[1]) == 0


def test_all_new_txns():
    ledger = Ledger(os.path.join("fixtures", "empty.lgr"))
    acct = Mock()
    acct.download = Mock(
        return_value=open(os.path.join("fixtures", "checking.ofx"), "rb")
    )
    sync = OfxSynchronizer(ledger)
    assert len(sync.get_new_txns(acct, 7, 7)[1]) == 3


def test_comment_txns():
    ledger = Ledger(os.path.join("fixtures", "empty.lgr"))
    sync = OfxSynchronizer(ledger)
    ofx = OfxSynchronizer.parse_file(os.path.join("fixtures", "comments.ofx"))
    txns = sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert len(txns) == 1


def test_sync_no_ledger():
    acct = Mock()
    acct.download = Mock(
        return_value=open(os.path.join("fixtures", "checking.ofx"), "rb")
    )
    sync = OfxSynchronizer(None)
    assert len(sync.get_new_txns(acct, 7, 7)[1]) == 3


def test_paypal_fresh_sync():
    ledger = Ledger(os.path.join("fixtures", "empty.lgr"))
    sync = CsvSynchronizer(ledger)
    assert 2 == len(sync.parse_file(os.path.join("fixtures", "paypal.csv")))


def test_sync_paypal_no_ledger():
    sync = CsvSynchronizer(None)
    assert 2 == len(sync.parse_file(os.path.join("fixtures", "paypal.csv")))


def test_paypal_partial_sync():
    ledger = Ledger(os.path.join("fixtures", "paypal.lgr"))
    sync = CsvSynchronizer(ledger)
    assert 1 == len(sync.parse_file(os.path.join("fixtures", "paypal.csv")))



# Copyright (c) 2013-2021 Erik Hetzner
# Portions Copyright (c) 2016 James S Blachly, MD
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import datetime
import hashlib
import re
from decimal import Decimal

from ofxparse.ofxparse import InvestmentTransaction
from ofxparse.ofxparse import Transaction as OfxTransaction

from ledgerautosync import EmptyInstitutionException

AUTOSYNC_INITIAL = "autosync_initial"
ALL_AUTOSYNC_INITIAL = "all.%s" % (AUTOSYNC_INITIAL)
UNKNOWN_BANK_ACCOUNT = "Assets:Unknown"


class EasyEquality(object):
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class SecurityList(object):
    """
    The SecurityList represents the OFX <SECLIST>...</SECLIST>
    and holds securities present in the OFX records

    <SECINFO>...</SECINFO> as implemented by OFXparse only includes:
    {memo, name, ticker, uniqueid}
    Unfortunately does not provide uniqueid_type or currency

    It is iterable, and also provides lookup table (LUT) functionality
    provides __next__() for Py3
    """

    def __init__(self, ofx):
        securities = []
        if hasattr(ofx, "security_list") and ofx.security_list is not None:
            securities = ofx.security_list

        self.cusip_lut = dict()
        self.ticker_lut = dict()

        self._iter = iter(securities)
        self.securities = securities
        if len(securities) == 0:
            return

        # index
        for sec in securities:
            # unfortunately OFXparse does not currently implement
            # security.uniqueid_type so I am presuming here
            if sec.uniqueid:
                self.cusip_lut[sec.uniqueid] = sec
            if sec.ticker:
                self.ticker_lut[sec.ticker] = sec
            # This indexing strategy (whereby I index the object instead of
            # the inverse value (e.g. ticker symbol) directly has a flaw
            # in that an OFX file could define a security list section and
            # list CUSIPs without ticker property, or the converse

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)

    def __len__(self):
        return len(self.securities)

    # one possibility is to just implement __getitem__(),
    # however since OFXparse does not implement securitylist.security.uniqueid_type
    # I'll have no idea if what I am seeing is a CUSISP
    # unless I look it up specifically as a CUSIP (and it exists)
    def find_cusip(self, cusip):
        if cusip in self.cusip_lut:
            return self.cusip_lut[cusip]
        else:
            return None

    def find_ticker(self, ticker):
        if ticker in self.ticker_lut:
            return self.ticker_lut[ticker]
        else:
            return None


class Transaction(object):
    def __init__(
        self,
        date,
        payee,
        postings,
        checknum=None,
        cleared=False,
        metadata={},
        aux_date=None,
        date_format=None,
    ):
        self.date = date
        self.aux_date = aux_date
        self.payee = payee
        self.postings = postings
        self.metadata = metadata
        self.cleared = cleared
        self.date_format = date_format
        self.checknum = checknum

    def format(self, indent=4, assertions=True):
        retval = ""
        cleared_str = " "
        checknum_str = ""
        if self.cleared:
            cleared_str = " * "
        aux_date_str = ""
        if self.date_format is None:
            self.date_format = "%Y/%m/%d"
        if self.aux_date is not None:
            aux_date_str = "=%s" % (self.aux_date.strftime(self.date_format))
        if self.checknum is not None:
            checknum_str = " (%d)" % (self.checknum)
        retval += "%s%s%s%s%s\n" % (
            self.date.strftime(self.date_format),
            checknum_str,
            aux_date_str,
            cleared_str,
            self.payee,
        )
        for k in sorted(self.metadata.keys()):
            retval += "%s; %s: %s\n" % (" " * indent, k, self.metadata[k])
        for posting in self.postings:
            retval += posting.format(indent, assertions)
        return retval


class Posting(object):
    def __init__(self, account, amount, asserted=None, unit_price=None, metadata={}):
        self.account = account
        self.amount = amount
        self.asserted = asserted
        self.unit_price = unit_price
        self.metadata = metadata

    def format(self, indent=4, assertions=True):
        space_count = 65 - indent - len(self.account) - len(self.amount.format())
        if space_count < 2:
            space_count = 2
        retval = "%s%s%s%s" % (
            " " * indent,
            self.account,
            " " * space_count,
            self.amount.format(),
        )
        if assertions and self.asserted is not None:
            retval = "%s = %s" % (retval, self.asserted.format())
        if self.unit_price is not None:
            retval = "%s @ %s" % (retval, self.unit_price.format())
        retval += "\n"
        for k in sorted(self.metadata.keys()):
            retval += "%s; %s: %s\n" % (" " * indent, k, self.metadata[k])
        return retval

    def clone_inverted(self, account, asserted=None, metadata={}):
        return Posting(
            account,
            self.amount.clone_inverted(),
            asserted=asserted,
            unit_price=self.unit_price,
            metadata=metadata,
        )


class Amount(EasyEquality):
    def __init__(self, number, currency, reverse=False, unlimited=False):
        self.number = Decimal(number)
        self.reverse = reverse
        self.unlimited = unlimited
        self.currency = currency

    def format(self):
        # Commodities must be quoted in ledger if they have
        # whitespace or numerals.
        if re.search(r"[\s0-9]", self.currency):
            currency = '"%s"' % (self.currency)
        else:
            currency = self.currency
        if self.unlimited:
            number = str(abs(self.number))
        else:
            number = "%0.2f" % (abs(self.number))
        if self.number.is_signed() != self.reverse:
            prefix = "-"
        else:
            prefix = ""
        if len(currency) == 1:
            # $ comes before
            return "%s%s%s" % (prefix, currency, number)
        else:
            # USD comes after
            return "%s%s %s" % (prefix, number, currency)

    def clone_inverted(self):
        return Amount(
            self.number,
            self.currency,
            reverse=not (self.reverse),
            unlimited=self.unlimited,
        )


class Converter(object):
    @staticmethod
    def clean_id(id):
        return (
            id.replace("/", "_")
            .replace("$", "_")
            .replace(" ", "_")
            .replace("@", "_")
            .replace("*", "_")
            .replace("+", "_")
            .replace("&", "_")
            .replace("[", "_")
            .replace("]", "_")
            .replace("|", "_")
            .replace("%", "_")
        )

    def __init__(
        self,
        ledger=None,
        unknownaccount=None,
        currency="$",
        indent=4,
        payee_format=None,
        date_format=None,
        infer_account=True,
    ):
        self.lgr = ledger
        self.indent = indent
        self.unknownaccount = unknownaccount
        self.currency = currency.upper()
        self.payee_format = payee_format
        if self.currency == "USD":
            self.currency = "$"
        self.date_format = date_format
        self.infer_account = infer_account

    def mk_dynamic_account(self, payee, exclude):
        if self.lgr is None or not self.infer_account:
            return self.unknownaccount or "Expenses:Misc"
        else:
            account = self.lgr.get_account_by_payee(payee, exclude)
            if account is None:
                return self.unknownaccount or "Expenses:Misc"
            else:
                return account


class OfxConverter(Converter):
    def __init__(
        self,
        account,
        name,
        indent=4,
        ledger=None,
        fid=None,
        unknownaccount=None,
        payee_format=None,
        hardcodeaccount=None,
        shortenaccount=False,
        security_list=SecurityList([]),
        date_format=None,
        infer_account=True,
    ):
        super(OfxConverter, self).__init__(
            ledger=ledger,
            indent=indent,
            unknownaccount=unknownaccount,
            currency=account.statement.currency,
            payee_format=payee_format,
            date_format=date_format,
            infer_account=infer_account,
        )
        self.real_acctid = account.account_id
        if hardcodeaccount is not None:
            self.acctid = hardcodeaccount
        elif shortenaccount:
            self.acctid = account.account_id[-4:]
        else:
            self.acctid = account.account_id
        self.payee_format = payee_format
        self.security_list = security_list

        if fid is not None:
            self.fid = fid
        else:
            if account.institution is None:
                raise EmptyInstitutionException(
                    "Institution provided by OFX is empty and no fid supplied!"
                )
            else:
                self.fid = account.institution.fid
        self.name = name

    def mk_ofxid(self, txnid):
        if self.acctid != self.real_acctid:
            # Some banks insert the bank account number into the transaction ID
            # We will do this to properly hide the account number for privacy
            # reasons
            txnid = txnid.replace(self.real_acctid, self.acctid)
        return Converter.clean_id("%s.%s.%s" % (self.fid, self.acctid, txnid))

    def format_payee(self, txn):
        payee = ""
        if hasattr(txn, "payee") and txn.payee is not None:
            payee = txn.payee
        memo = ""
        if hasattr(txn, "memo") and txn.memo is not None:
            memo = txn.memo
        txntype = ""
        if hasattr(txn, "type") and txn.type is not None:
            txntype = txn.type
        tferaction = ""
        if hasattr(txn, "tferaction") and txn.tferaction is not None:
            tferaction = txn.tferaction.lower()

        if payee != "" and self.lgr is not None:
            payee = self.lgr.get_autosync_payee(payee, self.name)

        # If payee is blank but memo is not, then try matching Ledger payee
        # based on memo instead.
        if memo != "" and payee == "" and self.lgr is not None:
            memo = self.lgr.get_autosync_payee(memo, self.name)

        payee_format = self.payee_format

        if payee_format is None:
            # Default when not provided
            payee_format = "{payee} {memo}"

            # Alternate for when payee and memo are blank, sometimes true for
            # investment accounts.
            if (payee == "") and (memo == ""):
                payee_format = "{account}: {txntype}"
                if tferaction != "":
                    payee_format += ": {tferaction}"

            # Sometimes memo/payee are simply longer versions of the other.
            if memo.startswith(payee):
                payee = ""
            if payee.startswith(memo):
                memo = ""

        return payee_format.format(
            payee=payee,
            memo=memo,
            txntype=txntype,
            account=self.name,
            tferaction=tferaction,
        ).strip()

    def format_balance(self, statement):
        # Get date. Ensure the date is a date-like object.
        if hasattr(statement, "balance_date") and hasattr(
            statement.balance_date, "strftime"
        ):
            date = statement.balance_date
        elif hasattr(statement, "end_date") and hasattr(statement.end_date, "strftime"):
            date = statement.end_date
        else:
            return ""
        if hasattr(statement, "balance"):
            return Transaction(
                date=date,
                cleared=True,
                payee="--Autosync Balance Assertion",
                postings=[
                    Posting(
                        self.name,
                        Amount(Decimal("0"), currency=self.currency),
                        asserted=Amount(statement.balance, self.currency),
                    )
                ],
                date_format=self.date_format,
            ).format(self.indent)
        else:
            return ""

    def format_initial_balance(self, statement):
        if hasattr(statement, "balance"):
            initbal = statement.balance
            for txn in statement.transactions:
                initbal -= txn.amount

            posting = Posting(
                self.name,
                Amount(initbal, currency=self.currency),
                metadata={"ofxid": self.mk_ofxid(AUTOSYNC_INITIAL)},
            )

            return Transaction(
                date=statement.start_date,
                payee="--Autosync Initial Balance",
                cleared=True,
                postings=[
                    posting,
                    posting.clone_inverted("Assets:Equity").format(self.indent),
                ],
                date_format=self.date_format,
            ).format(self.indent)
        else:
            return ""

    # Return the ticker symbol of the security with CUSIP, if it exists in the
    # security_list mapping. Otherwise, simply return the CUSIP.
    def maybe_get_ticker(self, cusip):
        security = self.security_list.find_cusip(cusip)
        if security and security.ticker:
            return security.ticker
        else:
            return cusip

    def convert(self, txn):
        """
        Convert an OFX Transaction to a Transaction
        """

        ofxid = self.mk_ofxid(txn.id)
        metadata = {}
        posting_metadata = {"ofxid": ofxid}

        if isinstance(txn, OfxTransaction):
            posting = Posting(
                self.name, Amount(txn.amount, self.currency), metadata=posting_metadata
            )
            checknum = None
            if hasattr(txn, "checknum") and txn.checknum != "":
                checknum = int(txn.checknum)
            return Transaction(
                date=txn.date,
                payee=self.format_payee(txn),
                postings=[
                    posting,
                    posting.clone_inverted(
                        self.mk_dynamic_account(
                            self.format_payee(txn), exclude=self.name
                        )
                    ),
                ],
                date_format=self.date_format,
                checknum=checknum,
            )
        elif isinstance(txn, InvestmentTransaction):
            acct1 = self.name
            acct2 = self.name

            posting1 = None
            posting2 = None

            security = self.maybe_get_ticker(txn.security)

            if isinstance(txn.type, str):
                # recent versions of ofxparse
                if re.match("^(buy|sell)", txn.type):
                    acct2 = self.unknownaccount or "Assets:Unknown"
                elif txn.type == "transfer":
                    acct2 = "Transfer"
                elif txn.type == "reinvest":
                    # reinvestment of income
                    # TODO: make this configurable
                    acct2 = "Income:Interest"
                elif txn.type == "income" and txn.income_type == "DIV":
                    # Fidelity lists non-reinvested dividend income as
                    # type: income, income_type: DIV
                    # TODO: determine how dividend income is listed from other institutions
                    # income/DIV transactions do not involve buying or selling a security
                    # so their postings need special handling compared to
                    # others
                    metadata["dividend_from"] = security
                    acct2 = "Income:Dividends"
                    posting1 = Posting(
                        acct1,
                        Amount(txn.total, self.currency),
                        metadata=posting_metadata,
                    )
                    posting2 = posting1.clone_inverted(acct2)
                else:
                    # ???
                    pass
            else:
                # Old version of ofxparse
                if txn.type in [0, 1, 3, 4]:
                    # buymf, sellmf, buystock, sellstock
                    acct2 = self.unknownaccount or "Assets:Unknown"
                elif txn.type == 2:
                    # reinvest
                    acct2 = "Income:Interest"
                else:
                    # ???
                    pass

            aux_date = None
            if txn.settleDate is not None and txn.settleDate != txn.tradeDate:
                aux_date = txn.settleDate

            # income/DIV already defined above;
            # this block defines all other posting types
            if posting1 is None and posting2 is None:
                posting_list = self._finalize_postings(
                    acct1, acct2, txn, security, posting_metadata
                )
            else:
                # Previously defined if type:income income_type/DIV
                posting_list = [posting1, posting2]

            return Transaction(
                date=txn.tradeDate,
                aux_date=aux_date,
                payee=self.format_payee(txn),
                metadata=metadata,
                postings=posting_list,
                date_format=self.date_format,
            )

    def _finalize_postings(self, acct1, acct2, txn, security, posting_metadata):
        """Finalize postings to include fees and other post-processing

        :param acct1:  Account for first posting.

        :param acct2:  Account for second posting.

        :param txn:    Instance of Transaction we are working on.

        :param security: security element we are working on.

        :param posting_metadata:  Posting metadata

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A list of Posting instances for txn.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method finalizes the postings for txn. We need
                  this because sometimes the postings are more complicated
                  (e.g., if we have fees or commissions for a stock trade).
                  See https://github.com/egh/ledger-autosync/issues/109.
        """
        posting_list = []
        posting_list.append(
            Posting(
                acct1,
                Amount(txn.units, security, unlimited=True),
                unit_price=Amount(txn.unit_price, self.currency, unlimited=True),
                metadata=posting_metadata,
            )
        )
        posting_list.append(
            Posting(
                acct2,
                Amount(
                    txn.units * txn.unit_price + txn.fees, self.currency, reverse=True
                ),
            )
        )
        if txn.fees:
            posting_list.append(
                Posting("Expenses:Fees", Amount(txn.fees, self.currency))
            )
        if txn.commission:
            posting_list.append(
                Posting("Expenses:Commission", Amount(txn.commission, self.currency))
            )

        return posting_list

    def format_position(self, pos):
        if (
            hasattr(pos, "date")
            and hasattr(pos, "security")
            and hasattr(pos, "unit_price")
        ):
            dateStr = pos.date.strftime("%Y/%m/%d %H:%M:%S")
            return "P %s %s %s\n" % (
                dateStr,
                self.maybe_get_ticker(pos.security),
                pos.unit_price,
            )


class CsvConverter(Converter):
    @staticmethod
    def make_converter(fieldset, dialect, name=None, **kwargs):
        for klass in CsvConverter.descendants():
            if klass.FIELDSET <= fieldset:
                return klass(dialect, name=name, **kwargs)
        # Found no class, bail
        raise Exception("Cannot determine CSV type")

    @classmethod
    def descendants(cls):
        retval = cls.__subclasses__()
        for cls2 in cls.__subclasses__():
            retval.extend(cls2.descendants())
        return retval

    # By default, return an MD5 of the key-value pairs in the row.
    # If a better ID is available, should be overridden.
    def get_csv_id(self, row):
        h = hashlib.md5()
        for key in sorted(row.keys()):
            h.update(("%s=%s\n" % (key, row[key])).encode("utf-8"))
        return h.hexdigest()

    def __init__(
        self,
        dialect,
        name=None,
        indent=4,
        ledger=None,
        unknownaccount=None,
        payee_format=None,
        date_format=None,
        infer_account=True,
    ):
        super(CsvConverter, self).__init__(
            ledger=ledger,
            indent=indent,
            unknownaccount=unknownaccount,
            payee_format=payee_format,
            date_format=date_format,
            infer_account=infer_account,
        )
        self.name = name
        self.dialect = dialect

    def format_payee(self, row):
        return re.sub(r"\s+", " ", self.payee_format.format(**row).strip())


class PaypalConverter(CsvConverter):
    FIELDSET = {
        "Currency",
        "Date",
        "Gross",
        "Item Title",
        "Name",
        "Net",
        "Status",
        "To Email Address",
        "Transaction ID",
        "Type",
    }

    def __init__(self, *args, **kwargs):
        super(PaypalConverter, self).__init__(*args, **kwargs)
        if self.payee_format is None:
            self.payee_format = (
                "{Name} {To Email Address} {Item Title} ID: {Transaction ID}, {Type}"
            )

    def get_csv_id(self, row):
        return "paypal.%s" % (Converter.clean_id(row["Transaction ID"]))

    def convert(self, row):
        if (
            (row["Status"] != "Completed")
            and (row["Status"] != "Refunded")
            and (row["Status"] != "Reversed")
        ) or (row["Type"] == "Shopping Cart Item"):
            return ""
        else:
            currency = row["Currency"]
            posting_metadata = {"csvid": self.get_csv_id(row)}
            net = Decimal(row["Net"].replace(",", ""))
            gross = Decimal(row["Gross"].replace(",", ""))

            if (
                row["Type"] == "Add Funds from a Bank Account"
                or row["Type"] == "Charge From Debit Card"
            ):
                posting = Posting(
                    self.name, Amount(net, currency), metadata=posting_metadata
                )
                postings = [posting, posting.clone_inverted("Transfer:Paypal")]
            else:
                posting = Posting(
                    self.name, Amount(gross, currency), metadata=posting_metadata
                )
                postings = [
                    posting,
                    # TODO Our payees are breaking the payee search in
                    # mk_dynamic_account
                    # self.mk_dynamic_account(payee, exclude=self.name),
                    posting.clone_inverted("Expenses:Misc"),
                ]
            return Transaction(
                date=datetime.datetime.strptime(row["Date"], "%m/%d/%Y"),
                payee=self.format_payee(row),
                postings=postings,
                date_format=self.date_format,
            )


# Apparently Paypal has another CSV


class PaypalAlternateConverter(CsvConverter):
    FIELDSET = {"Date", "Name", "Type", "Status", "Amount"}

    def __init__(self, *args, **kwargs):
        super(PaypalAlternateConverter, self).__init__(*args, **kwargs)
        if self.payee_format is None:
            self.payee_format = "{Name}: {Type}"

    def mk_amount(self, row, reverse=False):
        currency = "$"
        if "Currency" in row:
            currency = row["Currency"]
        return Amount(
            Decimal(re.sub(r"\$", "", row["Amount"])), currency, reverse=reverse
        )

    def convert(self, row):
        if (
            (row["Status"] != "Completed")
            and (row["Status"] != "Refunded")
            and (row["Status"] != "Reversed")
        ) or (row["Type"] == "Shopping Cart Item"):
            return ""
        else:
            posting_metadata = {"csvid": self.get_csv_id(row)}
            posting = Posting(self.name, self.mk_amount(row), metadata=posting_metadata)
            if (
                row["Type"] == "Add Funds from a Bank Account"
                or row["Type"] == "Charge From Debit Card"
            ):

                posting2_account = "Transfer:Paypal"
            else:
                posting2_account = "Expenses:Misc"
            return Transaction(
                date=datetime.datetime.strptime(row["Date"], "%m/%d/%Y"),
                payee=self.format_payee(row),
                postings=[posting, posting.clone_inverted(posting2_account)],
                date_format=self.date_format,
            )


class AmazonConverter(CsvConverter):
    FIELDSET = {"Currency", "Title", "Order Date", "Order ID"}

    def __init__(self, *args, **kwargs):
        super(AmazonConverter, self).__init__(*args, **kwargs)
        self.dialect.doublequote = True

    def mk_amount(self, row, reverse=False):
        currency = row["Currency"]
        if currency == "USD":
            currency = "$"
        return Amount(
            Decimal(re.sub(r"\$", "", row["Item Total"])), currency, reverse=reverse
        )

    def get_csv_id(self, row):
        return "amazon.%s" % (Converter.clean_id(row["Order ID"]))

    def convert(self, row):
        posting = Posting(
            self.name,
            self.mk_amount(row),
            metadata={
                "url": "https://www.amazon.com/gp/css/summary/print.html/ref=od_aui_print_invoice?ie=UTF8&orderID=%s"
                % (row["Order ID"]),  # noqa E501
                "csvid": self.get_csv_id(row),
            },
        )

        return Transaction(
            date=datetime.datetime.strptime(row["Order Date"], "%m/%d/%y"),
            payee=row["Title"],
            postings=[posting, posting.clone_inverted("Expenses:Misc")],
            date_format=self.date_format,
        )


class MintConverter(CsvConverter):
    FIELDSET = {
        "Date",
        "Amount",
        "Description",
        "Account Name",
        "Category",
        "Transaction Type",
    }

    def __init__(self, *args, **kwargs):
        super(MintConverter, self).__init__(*args, **kwargs)

    def mk_amount(self, row, reverse=False):
        return Amount(Decimal(row["Amount"]), "$", reverse=reverse)

    def convert(self, row):
        account = self.name
        if account is None:
            account = row["Account Name"]
        postings = []
        posting_metadata = {"csvid": "mint.%s" % (self.get_csv_id(row))}
        if row["Transaction Type"] == "credit":
            posting = Posting(
                account, self.mk_amount(row, reverse=True), metadata=posting_metadata
            )
            postings = [posting, posting.clone_inverted(row["Category"])]
        else:
            posting = Posting(account, self.mk_amount(row), metadata=posting_metadata)
            postings = [
                posting,
                posting.clone_inverted("Expenses:%s" % (row["Category"])),
            ]

        return Transaction(
            date=datetime.datetime.strptime(row["Date"], "%m/%d/%Y"),
            payee=row["Description"],
            postings=postings,
            date_format=self.date_format,
        )


# Simple.com
class SimpleConverter(CsvConverter):
    FIELDSET = {
        "Date",
        "Recorded at",
        "Scheduled for",
        "Amount",
        "Activity",
        "Pending",
        "Raw description",
        "Description",
        "Category folder",
        "Category",
        "Street address",
        "City",
        "State",
        "Zip",
        "Latitude",
        "Longitude",
        "Memo",
    }

    def __init__(self, *args, **kwargs):
        super(SimpleConverter, self).__init__(*args, **kwargs)

    def convert(self, row):
        amount = abs(float(row["Amount"]))
        reverse = row["Amount"][0] == "-"

        if reverse:
            account = "Expenses:%s" % (row["Category"])
        else:
            account = "Income:%s" % (row["Category"])

        posting_metadata = {
            "csvid": "simple.%s" % (self.get_csv_id(row)),
            "raw_description": row["Raw description"],
            "activity_type": row["Activity"],
        }
        if row["Memo"]:
            posting_metadata["memo"] = row["Memo"]

        return Transaction(
            date=datetime.datetime.strptime(row["Date"], "%Y/%m/%d"),
            payee=row["Description"],
            postings=[
                Posting(
                    self.name, Amount(amount, "$", reverse), metadata=posting_metadata
                ),
                Posting(account, Amount(amount, "$", reverse=not (reverse))),
            ],
            date_format=self.date_format,
        )


class VenmoConverter(CsvConverter):
    FIELDSET = set(
        [
            "Username",
            "ID",
            "Datetime",
            "Type",
            "Status",
            "Note",
            "From",
            "To",
            "Amount (total)",
            "Amount (fee)",
            "Funding Source",
            "Destination",
            "Beginning Balance",
            "Ending Balance",
            "Statement Period Venmo Fees",
            "Year to Date Venmo Fees",
            "Disclaimer",
        ]
    )

    def __init__(self, *args, **kwargs):
        # Initialize max date to be used for balance assertion
        self.max_date = datetime.datetime.min
        super(VenmoConverter, self).__init__(*args, **kwargs)

    def convert(self, row):

        # Venmo has a hierarchical structure with some rows containing
        # statement level data
        if not row["ID"]:
            if row["Ending Balance"]:
                # This relies `self.max_date` being correct, which in turn
                # relies on the statement balance being after all transactions.
                # This seems to be how the statement is currently formatted but
                # may break if Venmo changes format
                return Transaction(
                    date=self.max_date,
                    cleared=True,
                    payee="--Autosync Balance Assertion",
                    postings=[
                        Posting(
                            self.name,
                            Amount(Decimal("0"), currency=self.currency),
                            asserted=Amount(
                                row["Ending Balance"].replace("$", "").replace(",", ""),
                                self.currency,
                            ),
                        )
                    ],
                    date_format=self.date_format,
                ).format(self.indent)
            else:
                return ""

        md = re.match(r"^([+-]) \$([0-9,\.]+)", row["Amount (total)"])
        amount = md.group(2).replace(",", "")
        if md.group(1) == "-":
            reverse = True
            account = self.unknownaccount or "expenses"
        else:
            reverse = False
            account = self.unknownaccount or "income"

        if row["Type"] == "Payment":
            if reverse:
                payee = row["To"]
            else:
                payee = row["From"]
        else:
            if reverse:
                payee = row["From"]
            else:
                payee = row["To"]
        date = datetime.datetime.strptime(row["Datetime"], "%Y-%m-%dT%H:%M:%S")
        if date > self.max_date:
            self.max_date = date

        return Transaction(
            date=date,
            payee=payee,
            metadata={"desc": row["Note"]},
            postings=[
                Posting(
                    self.name,
                    Amount(amount, "$", reverse=reverse),
                    metadata={"csvid": self.get_csv_id(row)},
                ),
                Posting(account, Amount(amount, "$", reverse=not (reverse))),
            ],
            date_format=self.date_format,
        )

    def get_csv_id(self, row):
        return "venmo.{}".format(Converter.clean_id(row["ID"]))



# ledger-autosync plugin for CSV files from First Direct, a UK bank.
# The currency is fixed to GBP for that reason.

import datetime
import re

from ledgerautosync.converter import Amount, CsvConverter, Posting, Transaction


class SomeConverter(CsvConverter):
    FIELDSET = set(["Date", "Description", "Amount"])

    def __init__(self, *args, **kwargs):
        super(SomeConverter, self).__init__(*args, **kwargs)

    def convert(self, row):
        amount = row["Amount"]
        return Transaction(
            date=datetime.datetime.strptime(row["Date"], "%d/%m/%Y"),
            payee=row["Description"].strip(),
            postings=[
                Posting(self.name, Amount(amount, "GBP")),
                Posting(self.unknownaccount, Amount(amount, "GBP", reverse=True)),
            ],
        )



# ledger-autosync plugin for CSV files from AIB, an Irish bank.

import datetime
from decimal import Decimal
import re

from ledgerautosync.converter import Amount, Converter, CsvConverter, Posting, Transaction

class AIBConverter(CsvConverter):
    FIELDSET = set(["Posted Account", "Posted Transactions Date", "Description",
        "Debit Amount", "Credit Amount", "Balance", "Transaction Type"])

    def __init__(self, *args, **kwargs):
        super(AIBConverter, self).__init__(*args, **kwargs)

    def convert(self, row):
        meta = {"csvid": self.get_csv_id(row)}
        debit  = Decimal(row["Debit Amount"])  if row["Debit Amount"]  else None
        credit = Decimal(row["Credit Amount"]) if row["Credit Amount"] else None
        if (debit and debit > 0):
            posting_from = Posting(self.name, Amount(debit, "€", reverse=True), metadata=meta)
            posting_to   = posting_from.clone_inverted("Expenses:Misc")
        elif (credit and credit > 0):
            posting_from = Posting("Assets:Other", Amount(credit, "€", reverse=True))
            posting_to   = posting_from.clone_inverted(self.name, metadata=meta)
        else:
            return ""

        payee = re.sub(r"[A-Za-z]+('[A-Za-z]+)?",lambda word: word.group(0).capitalize(),row["Description"])

        return Transaction(
            date=datetime.datetime.strptime(row["Posted Transactions Date"], "%d/%m/%y"),
            cleared=True,
            date_format="%Y-%m-%d",
            payee=payee,
            postings=[posting_to, posting_from]
        )



# ledger-autosync plugin for CSV files from Revolut, a Lithuania-based online bank.

import datetime
from decimal import Decimal
import re

from ledgerautosync.converter import Amount, Converter, CsvConverter, Posting, Transaction

class RevolutConverter(CsvConverter):
    FIELDSET = set(["Type","Product","Started Date","Completed Date",
        "Description","Amount","Fee","Currency","State","Balance"])

    def __init__(self, *args, **kwargs):
        super(RevolutConverter, self).__init__(*args, **kwargs)

    def mk_currency(self, currency):
        if currency == "USD":
            currency = "$"
        elif currency == "GBP":
            currency = "£"
        elif currency == "EUR":
            currency = "€"
        return currency

    def convert(self, row):
        amt = Decimal(row["Amount"])
        if not amt:
            return ""
        currency = self.mk_currency(row["Currency"])
        cleared = row["State"] == "COMPLETED"
        if row["Type"] == "TOPUP":
            reverse   = True
            acct_from = "Assets:Other"
            amt_from  = Amount(amt, currency, reverse=reverse)
            acct_to   = self.name
            amt_to    = Amount(amt, currency, reverse=not reverse)
        else:
            reverse   = False
            acct_from = self.name
            amt_from  = Amount(amt, currency, reverse=reverse)
            acct_to   = "Expenses:Misc"
            amt_to    = Amount(amt, currency, reverse=not reverse)

        payee = row["Description"]
        meta = {"csvid": self.get_csv_id(row)}

        posting_from = Posting(acct_from, amt_from, metadata=meta if acct_from == self.name else {})
        posting_fee  = Posting("Expenses:Bank Charges", Amount(Decimal(row["Fee"]), currency, reverse=True)) if row["Fee"] != "0.00" else None
        posting_to   = Posting(acct_to,   amt_to,   metadata=meta if acct_to   == self.name else {})

        date     = datetime.datetime.strptime(row["Started Date"], "%Y-%m-%d %H:%M:%S")
        aux_date = datetime.datetime.strptime(row["Completed Date"], "%Y-%m-%d %H:%M:%S") if row["Completed Date"] else None
        if aux_date and (date.date() == aux_date.date()):
            aux_date = None

        postings = [posting_to, posting_from]
        if posting_fee:
            postings.append(posting_fee)
            postings.append(posting_fee.clone_inverted(self.name))
        return Transaction(
            date=date,
            cleared=cleared,
            aux_date=aux_date,
            date_format="%Y-%m-%d",
            payee=payee,
            postings=postings
        )



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import os

import pytest
from ofxparse import OfxParser

from ledgerautosync.ledgerwrap import HLedger, Ledger, LedgerPython

LEDGER = [HLedger, Ledger, LedgerPython]


@pytest.fixture(params=LEDGER)
def ledger(request):
    lgr_name = request.node.get_closest_marker("lgr_file").args[0]
    ledger_impls = request.node.get_closest_marker("ledger_impls")
    kwargs = {}
    if request.param == Ledger:
        kwargs["no_pipe"] = True
    if not request.param.available():
        pytest.skip(f"{request.param} not found")
    if ledger_impls is not None:
        if request.param not in ledger_impls.args[0]:
            pytest.skip("Test not applicable for this ledger.")
    return request.param(os.path.join("fixtures", lgr_name), **kwargs)


@pytest.fixture
def ofx(request):
    ofx_name = request.node.get_closest_marker("ofx_file").args[0]
    with open(os.path.join("fixtures", ofx_name), "rb") as ofx_file:
        yield OfxParser.parse(ofx_file)



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

import re

import pytest

from ledgerautosync.converter import OfxConverter, SecurityList
from ledgerautosync.ledgerwrap import Ledger


def clean_posting(posting):
    return re.sub("  +", "  ", posting)


def make_clean_posting(converter, txn):
    return clean_posting(converter.convert(txn).format())


@pytest.mark.ofx_file("checking.ofx")
def test_checking(ofx):
    converter = OfxConverter(account=ofx.account, name="Foo")
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
  Foo  $0.01
  ; ofxid: 1101.1452687~7.0000486
  Expenses:Misc  -$0.01
"""
    )
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[1])
        == """2011/04/05 AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )
  Foo  -$34.51
  ; ofxid: 1101.1452687~7.0000487
  Expenses:Misc  $34.51
"""
    )

    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[2])
        == """2011/04/07 (319) RETURNED CHECK FEE, CHECK # 319 FOR $45.33 ON 04/07/11
  Foo  -$25.00
  ; ofxid: 1101.1452687~7.0000488
  Expenses:Misc  $25.00
"""
    )


@pytest.mark.ofx_file("checking.ofx")
def test_indent(ofx):
    converter = OfxConverter(account=ofx.account, name="Foo", indent=4)
    # testing indent, so do not use the string collapsing version of assert
    assert (
        converter.convert(ofx.account.statement.transactions[0]).format()
        == """2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
    Foo                                                     $0.01
    ; ofxid: 1101.1452687~7.0000486
    Expenses:Misc                                          -$0.01
"""
    )


@pytest.mark.ofx_file("checking.ofx")
def test_shortenaccount(ofx):
    converter = OfxConverter(
        account=ofx.account, name="Foo", indent=4, shortenaccount=True
    )
    # testing indent, so do not use the string collapsing version of assert
    assert (
        converter.convert(ofx.account.statement.transactions[0]).format()
        == """2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
    Foo                                                     $0.01
    ; ofxid: 1101.87~7.0000486
    Expenses:Misc                                          -$0.01
"""
    )


@pytest.mark.ofx_file("checking.ofx")
def test_hardcodeaccount(ofx):
    converter = OfxConverter(
        account=ofx.account, name="Foo", indent=4, hardcodeaccount="9999"
    )
    # testing indent, so do not use the string collapsing version of assert
    assert (
        converter.convert(ofx.account.statement.transactions[0]).format()
        == """2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
    Foo                                                     $0.01
    ; ofxid: 1101.9999.0000486
    Expenses:Misc                                          -$0.01
"""
    )


@pytest.mark.ofx_file("fidelity.ofx")
def test_investments(ofx):
    converter = OfxConverter(
        account=ofx.account,
        name="Foo",
        security_list=SecurityList(ofx),
    )
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2012/07/20 YOU BOUGHT
  Foo  100.00000 INTC @ $25.635000000
  ; ofxid: 7776.01234567890.0123456789020201120120720
  Assets:Unknown  -$2563.50
  Expenses:Commission  $7.95
"""
    )
    # test no payee/memo
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[1])
        == """2012/07/27 Foo: buystock
  Foo  128.00000 SDRL @ $39.390900000
  ; ofxid: 7776.01234567890.0123456789020901120120727
  Assets:Unknown  -$5042.04
  Expenses:Commission  $7.95
"""
    )


@pytest.mark.ofx_file("fidelity_fee.ofx")
def test_fee(ofx):
    """Test that fees are parsed correctly.

    In this case we have a 7-cent fee. We need to make sure that
    the net sale price which shows up is the gross price of 3239.44
    minus 7 cents which equals 3239.37 and that the 7 cent fee
    shows up as an extra posting.
    """
    converter = OfxConverter(
        account=ofx.account, name="Foo", security_list=SecurityList(ofx)
    )
    # test fee
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[1])
        == """2012/08/01 SELL
  Foo  -100.0 "929042109" @ $32.3944
  ; ofxid: 7776.01234567890.0123456789021401420120801
  Assets:Unknown  $3239.37
  Expenses:Fees  $0.07
"""
    )
    # test fee and comission
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2020/05/22=2020/05/26 SELL
  Foo  -1.0 "Z9977810Z" @ $8.27
  ; ofxid: 7776.01234567890.987654321
  Assets:Unknown  $8.25
  Expenses:Fees  $0.02
  Expenses:Commission  $1.00
"""
    )


@pytest.mark.ofx_file("checking.ofx")
@pytest.mark.lgr_file("checking-dynamic-account.lgr")
def test_dynamic_account(ofx, ledger):
    converter = OfxConverter(account=ofx.account, name="Assets:Foo", ledger=ledger)
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[1])
        == """2011/04/05 AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )
  Assets:Foo  -$34.51
  ; ofxid: 1101.1452687~7.0000487
  Expenses:Bar  $34.51
"""
    )


@pytest.mark.ofx_file("checking.ofx")
@pytest.mark.lgr_file("checking.lgr")
def test_balance_assertion(ofx, ledger):
    converter = OfxConverter(account=ofx.account, name="Assets:Foo", ledger=ledger)
    assert (
        clean_posting(converter.format_balance(ofx.account.statement))
        == """2013/05/25 * --Autosync Balance Assertion
  Assets:Foo  $0.00 = $100.99
"""
    )


@pytest.mark.ofx_file("checking.ofx")
@pytest.mark.lgr_file("checking.lgr")
def test_initial_balance(ofx, ledger):
    converter = OfxConverter(account=ofx.account, name="Assets:Foo", ledger=ledger)
    assert (
        clean_posting(converter.format_initial_balance(ofx.account.statement))
        == """2000/01/01 * --Autosync Initial Balance
  Assets:Foo  $160.49
  ; ofxid: 1101.1452687~7.autosync_initial
  Assets:Equity  -$160.49
"""
    )


@pytest.mark.ofx_file("checking.ofx")
def test_unknownaccount(ofx):
    converter = OfxConverter(
        account=ofx.account, name="Foo", unknownaccount="Expenses:Unknown"
    )
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
  Foo  $0.01
  ; ofxid: 1101.1452687~7.0000486
  Expenses:Unknown  -$0.01
"""
    )


@pytest.mark.ofx_file("fidelity.ofx")
def test_quote_commodity(ofx):
    converter = OfxConverter(
        account=ofx.account, name="Foo", security_list=SecurityList(ofx)
    )
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2012/07/20 YOU BOUGHT
  Foo  100.00000 INTC @ $25.635000000
  ; ofxid: 7776.01234567890.0123456789020201120120720
  Assets:Unknown  -$2563.50
  Expenses:Commission  $7.95
"""
    )


# Check that <TRANSFER> txns are parsed.
@pytest.mark.ofx_file("investment_401k.ofx")
def test_transfer_txn(ofx):
    converter = OfxConverter(
        account=ofx.account,
        name="Foo",
        unknownaccount="Expenses:Unknown",
    )
    if len(ofx.account.statement.transactions) > 2:
        # older versions of ofxparse would skip these transactions
        if hasattr(ofx.account.statement.transactions[2], "tferaction"):
            # unmerged pull request
            assert (
                make_clean_posting(converter, ofx.account.statement.transactions[2])
                == """2014/06/30 Foo: transfer: out
  Foo  -9.060702 BAZ @ $21.928764
  ; ofxid: 1234.12345678.123456-01.3
  Transfer  $198.69
"""
            )
        else:
            assert (
                make_clean_posting(converter, ofx.account.statement.transactions[2])
                == """2014/06/30 Foo: transfer
  Foo  -9.060702 BAZ @ $21.928764
  ; ofxid: 1234.12345678.123456-01.3
  Transfer  $198.69
"""
            )


@pytest.mark.ofx_file("cusip.ofx")
def test_position(ofx):
    converter = OfxConverter(
        account=ofx.account,
        name="Foo",
        indent=4,
        unknownaccount="Expenses:Unknown",
        security_list=SecurityList(ofx),
    )
    assert (
        converter.format_position(ofx.account.statement.positions[0])
        == """P 2016/10/08 07:30:08 SHSAX 47.8600000
"""
    )


@pytest.mark.ofx_file("income.ofx")
def test_dividend(ofx):
    converter = OfxConverter(account=ofx.account, name="Foo")
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2016/10/12 DIVIDEND RECEIVED
  ; dividend_from: cusip_redacted
  Foo  $1234.56
  ; ofxid: 1234.12345678.123456-01.redacted
  Income:Dividends  -$1234.56
"""
    )


@pytest.mark.ofx_file("checking.ofx")
def test_checking_custom_payee(ofx):
    converter = OfxConverter(account=ofx.account, name="Foo", payee_format="{memo}")
    assert (
        converter.format_payee(ofx.account.statement.transactions[0])
        == "DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%"
    )
    converter = OfxConverter(account=ofx.account, name="Foo", payee_format="{payee}")
    assert (
        converter.format_payee(ofx.account.statement.transactions[0])
        == "DIVIDEND EARNED FOR PERIOD OF 03"
    )
    converter = OfxConverter(account=ofx.account, name="Foo", payee_format="{account}")
    assert converter.format_payee(ofx.account.statement.transactions[0]) == "Foo"
    converter = OfxConverter(
        account=ofx.account, name="Foo", payee_format=" {account} "
    )
    assert converter.format_payee(ofx.account.statement.transactions[0]) == "Foo"


@pytest.mark.ofx_file("investment_401k.ofx")
def test_investments_custom_payee(ofx):
    converter = OfxConverter(account=ofx.account, name="Foo", payee_format="{txntype}")
    assert converter.format_payee(ofx.account.statement.transactions[1]) == "transfer"
    converter = OfxConverter(
        account=ofx.account, name="Foo", payee_format="{tferaction}"
    )
    assert converter.format_payee(ofx.account.statement.transactions[1]) == "in"


@pytest.mark.ofx_file("checking-payee-match.ofx")
@pytest.mark.lgr_file("checking.lgr")
@pytest.mark.ledger_impls([Ledger])
def test_payee_match(ofx, ledger):
    converter = OfxConverter(account=ofx.account, name="Foo", ledger=ledger)
    assert (
        make_clean_posting(converter, ofx.account.statement.transactions[0])
        == """2011/03/31 Match Payee
  Foo  -$0.01
  ; ofxid: 1101.1452687~7.0000489
  Expenses:Bar  $0.01
"""
    )



# ledger-autosync plugin for CSV files from Wise, formerly TransferWise.

# You must download and sync a CSV statement for all currencies that
# cover the same time period.  Since Wise lists the transactions in
# reverse chronological order, you might want to run with --reverse flag.
# The code assumes some text is in English.  Adjust mk_currency() and
# anything else to suit. No warranty, YMMV, etc.

import datetime
from decimal import Decimal
import re

from ledgerautosync.converter import Amount, Converter, CsvConverter, Posting, Transaction

class WiseConverter(CsvConverter):
    FIELDSET = set(["TransferWise ID","Date","Amount","Currency","Description","Payment Reference","Running Balance",
                    "Exchange From","Exchange To","Exchange Rate","Payer Name","Payee Name","Payee Account Number",
                    "Merchant","Card Last Four Digits","Card Holder Full Name","Attachment","Note","Total fees"])

    def __init__(self, *args, **kwargs):
        super(WiseConverter, self).__init__(*args, **kwargs)

    def mk_currency(self, currency):
        if currency == "USD":
            currency = "$"
        elif currency == "GBP":
            currency = "£"
        elif currency == "EUR":
            currency = "€"
        return currency

    def mk_amount(self, amt, currency, reverse=False):
        currency = self.mk_currency(currency)
        return Amount(
            Decimal(amt), currency, reverse=reverse
        )

    def convert(self, row):
        tid = row["TransferWise ID"]
        checknum = int(tid.split("-")[1])
        amt = Decimal(row["Amount"])
        acct_from = self.name
        curr_from = self.mk_currency(row["Currency"])
        acct_to   = "Expenses:Misc"
        curr_to   = curr_from
        amt_from  = Amount(amt, curr_from)
        amt_to    = Amount(amt, curr_to, reverse=True)

        fee_not_included = tid.startswith("CARD-") and row["Currency"] == "USD" and row["Exchange To"]

        if row["Exchange To"]:
            rate      = Decimal(row["Exchange Rate"])
            curr      = self.mk_currency(row["Currency"])
            curr_from = self.mk_currency(row["Exchange From"])
            curr_to   = self.mk_currency(row["Exchange To"])
            if curr == curr_from:
                amt_from  = Amount(amt,curr_from)
                # Card transactions from USD to other currencies do not consider the fees in the exchange rate
                amt_to    = Amount((amt + (Decimal(row["Total fees"]) if tid.startswith("CARD-") and not (row["Currency"] == "USD" and row["Exchange To"]) else Decimal(0)))  * rate, curr_to, reverse=True)
                acct_from = self.name
            else:
                # Do not import this exchange from this statement; instead use the statement for the matching "from" currency
                if tid.startswith("BALANCE-"):
                    return ""
                amt_from  = Amount(amt / rate, curr_from, reverse=True)
                amt_to     = Amount(amt, curr_to)
                acct_from = self.name if tid.startswith("BALANCE-") else "Expenses:Misc"
                if tid.startswith("BALANCE-"):
                    acct_to = self.name

        if row["Description"].startswith("Wise Charges for:"):
            acct_to = "Expenses:Bank Charges"

        if tid.startswith("TRANSFER-"):
            payee = row["Payee Name"] or row["Payer Name"]
        elif tid.startswith("BALANCE-"):
            payee = row["Description"]
        else:
            payee = row["Merchant"]

        meta = {"csvid": self.get_csv_id(row)}

        posting_from = Posting(acct_from, amt_from, metadata=meta)
        posting_to   = Posting(acct_to,   amt_to)

        return Transaction(
            date=datetime.datetime.strptime(row["Date"], "%d-%m-%Y"),
            cleared=True,
            date_format="%Y-%m-%d",
            checknum=checknum,
            payee=payee,
            postings=[posting_to, posting_from]
        )

    def get_csv_id(self, row):
        fmt = 'wise.fee.{}' if row["Description"].startswith("Wise Charges for:") else "wise.{}"
        return fmt.format(Converter.clean_id(row["TransferWise ID"]))



# -*- coding: utf-8 -*-
# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import os.path

import pytest
from ofxclient.config import OfxConfig

from ledgerautosync import EmptyInstitutionException
from ledgerautosync.cli import run
from ledgerautosync.converter import OfxConverter
from ledgerautosync.sync import OfxSynchronizer


@pytest.fixture
def ofx_sync(ledger):
    return OfxSynchronizer(ledger)


@pytest.mark.lgr_file("empty.lgr")
def test_no_institution_no_fid(ledger):
    with pytest.raises(EmptyInstitutionException):
        config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
        run(
            [
                os.path.join("fixtures", "no-institution.ofx"),
                "-l",
                os.path.join("fixtures", "empty.lgr"),
                "-a",
                "Assets:Savings:Foo",
            ],
            config,
        )


@pytest.mark.ofx_file("no-institution.ofx")
@pytest.mark.lgr_file("empty.lgr")
def test_no_institution(ofx, ofx_sync):
    txns = ofx_sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert len(txns) == 3


@pytest.mark.ofx_file("no-institution.ofx")
@pytest.mark.lgr_file("empty.lgr")
def test_no_institution_no_accountname(ofx):
    with pytest.raises(EmptyInstitutionException):
        OfxConverter(account=ofx.account, name=None)


@pytest.mark.ofx_file("apostrophe.ofx")
@pytest.mark.lgr_file("empty.lgr")
def test_apostrophe(ofx, ofx_sync):
    txns = ofx_sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert len(txns) == 1


@pytest.mark.ofx_file("fidelity-one-dtsettle.ofx")
@pytest.mark.lgr_file("empty.lgr")
def test_one_settleDate(ofx, ofx_sync):
    txns = ofx_sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    assert len(txns) == 17


@pytest.mark.ofx_file("accented_characters_latin1.ofx")
@pytest.mark.lgr_file("empty.lgr")
def test_accented_characters_latin1(ofx, ofx_sync):
    txns = ofx_sync.filter(ofx.account.statement.transactions, ofx.account.account_id)
    converter = OfxConverter(account=ofx.account, name="Foo")
    assert converter.format_payee(txns[0]) == "Virement Interac à: Jean"
    assert len(txns) == 1



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

import csv
import hashlib
import os
from decimal import Decimal

import pytest

from ledgerautosync.converter import (
    AmazonConverter,
    Amount,
    CsvConverter,
    MintConverter,
    PaypalAlternateConverter,
    PaypalConverter,
    Posting,
)


def test_format():
    assert (
        Posting("Foo", Amount(Decimal("10.00"), "$"), metadata={"foo": "bar"}).format(
            indent=2
        )
        == "  Foo                                                      $10.00\n  ; foo: bar\n"
    )


def test_amount():
    assert (
        "$10.00" == Amount(Decimal("10.001"), "$").format()
    ), "Formats to 2 points precision by default"
    assert (
        "10.00 USD" == Amount(Decimal(10), "USD").format()
    ), "Longer commodity names come after"
    assert (
        "-$10.00" == Amount(Decimal(10), "$", reverse=True).format()
    ), "Reverse flag works."
    assert (
        '10.00 "ABC123"' == Amount(Decimal(10), "ABC123").format()
    ), "Currencies with numbers are quoted"
    assert (
        '10.00 "A BC"' == Amount(Decimal(10), "A BC").format()
    ), "Currencies with whitespace are quoted"
    assert "$10.001" == Amount(Decimal("10.001"), "$", unlimited=True).format()


def test_get_csv_id():
    converter = CsvConverter(None)
    h = {"foo": "bar", "bar": "foo"}
    assert (
        converter.get_csv_id(h)
        == hashlib.md5("bar=foo\nfoo=bar\n".encode("utf-8")).hexdigest()
    )


@pytest.fixture
def reader(request):
    csv_name = request.node.get_closest_marker("csv_file").args[0]
    with open(os.path.join("fixtures", csv_name), "r") as csv_file:
        dialect = csv.Sniffer().sniff(csv_file.read(1024))
        csv_file.seek(0)
        dialect.skipinitialspace = True
        reader = csv.DictReader(csv_file, dialect=dialect)
        csv_file.seek(0)
        reader = csv.DictReader(csv_file, dialect=dialect)
        yield reader


@pytest.fixture
def converter(reader):
    return CsvConverter.make_converter(set(reader.fieldnames), reader.dialect, "Foo")


@pytest.mark.csv_file("paypal.csv")
def test_paypal_format(reader, converter):
    assert type(converter) == PaypalConverter
    assert (
        converter.convert(next(reader)).format()
        == """2016/06/04 Jane Doe someone@example.net My Friend ID: XYZ1, Recurring Payment Sent
    Foo                                                -20.00 USD
    ; csvid: paypal.XYZ1
    Expenses:Misc                                       20.00 USD
"""
    )
    assert (
        converter.convert(next(reader)).format()
        == """2016/06/04 Debit Card ID: XYZ2, Charge From Debit Card
    Foo                                               1120.00 USD
    ; csvid: paypal.XYZ2
    Transfer:Paypal                                  -1120.00 USD
"""
    )


@pytest.mark.csv_file("paypal_alternate.csv")
def test_paypal_alternate_format(reader, converter):
    assert type(converter) == PaypalAlternateConverter
    assert (
        converter.convert(next(reader)).format()
        == """2016/12/31 Some User: Payment Sent
    Foo                                                   -$12.34
    ; csvid: 1209a7bb0d17276248d463b71a6a8b8c
    Expenses:Misc                                          $12.34
"""
    )
    assert (
        converter.convert(next(reader)).format()
        == """2016/12/31 Bank Account: Add Funds from a Bank Account
    Foo                                                    $12.34
    ; csvid: 581e62da71bab74c7ce61854c2b6b6a5
    Transfer:Paypal                                       -$12.34
"""
    )


def test_mk_amount_alternate():
    converter = PaypalAlternateConverter(None)
    row = {"Currency": "USD", "Amount": "12.34"}
    assert converter.mk_amount(row) == Amount(Decimal("12.34"), "USD")


@pytest.mark.csv_file("amazon.csv")
def test_amazon_format(reader, converter):
    assert type(converter) == AmazonConverter
    assert (
        converter.convert(next(reader)).format()
        == """2016/01/29 Best Soap Ever
    Foo                                                    $21.90
    ; csvid: amazon.123-4567890-1234567
    ; url: https://www.amazon.com/gp/css/summary/print.html/ref=od_aui_print_invoice?ie=UTF8&orderID=123-4567890-1234567
    Expenses:Misc                                         -$21.90
"""
    )


@pytest.mark.csv_file("amazon2.csv")
def test_amazon2_format(reader, converter):
    assert type(converter) == AmazonConverter
    assert (
        converter.convert(next(reader)).format()
        == """2017/06/05 Test " double quote
    Foo                                                     $9.99
    ; csvid: amazon.111-1111111-1111111
    ; url: https://www.amazon.com/gp/css/summary/print.html/ref=od_aui_print_invoice?ie=UTF8&orderID=111-1111111-1111111
    Expenses:Misc                                          -$9.99
"""
    )


@pytest.mark.csv_file("mint.csv")
def test_mint_format(reader):
    converter = CsvConverter.make_converter(set(reader.fieldnames), reader.dialect)
    assert type(converter) == MintConverter
    assert (
        converter.convert(next(reader)).format()
        == """2016/08/02 Amazon
    1234                                                   $29.99
    ; csvid: mint.a7c028a73d76956453dab634e8e5bdc1
    Expenses:Shopping                                     -$29.99
"""
    )
    assert (
        converter.convert(next(reader)).format()
        == """2016/06/02 Autopay Rautopay Auto
    1234                                                 -$123.45
    ; csvid: mint.a404e70594502dd62bfc6f15d80b7cd7
    Credit Card Payment                                   $123.45
"""
    )



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import codecs
import csv
import logging

from ofxparse import OfxParser, OfxParserException

from ledgerautosync.converter import CsvConverter


class Synchronizer(object):
    def __init__(self, lgr):
        self.lgr = lgr


class OfxSynchronizer(Synchronizer):
    def __init__(self, lgr, hardcodeaccount=None, shortenaccount=None):
        self.hardcodeaccount = hardcodeaccount
        self.shortenaccount = shortenaccount
        super(OfxSynchronizer, self).__init__(lgr)

    @staticmethod
    def parse_file(path):
        with open(path, "rb") as ofx_file:
            return OfxParser.parse(ofx_file)

    def is_txn_synced(self, acctid, txn):
        if self.lgr is None:
            # User called with --no-ledger
            # All transactions are considered "synced" in this case.
            return False
        else:
            acctid_to_use = acctid
            txnid_to_use = txn.id
            if self.hardcodeaccount:
                acctid_to_use = self.hardcodeaccount
                txnid_to_use = txnid_to_use.replace(acctid, acctid_to_use)
            elif self.shortenaccount:
                acctid_to_use = acctid[-4:]
                txnid_to_use = txnid_to_use.replace(acctid, acctid_to_use)
            ofxid = "%s.%s" % (acctid_to_use, txnid_to_use)
            return self.lgr.check_transaction_by_id("ofxid", ofxid)

    # Filter out comment transactions. These have an amount of 0 and the same
    # datetime as the previous transactions.
    def filter_comment_txns(self, txns):
        last_txn = None
        retval = []
        for txn in txns:
            if (
                (last_txn is not None)
                and hasattr(txn, "amount")
                and (txn.amount == 0)
                and hasattr(last_txn, "date")
                and hasattr(txn, "date")
                and (last_txn.date == txn.date)
            ):
                # This is a comment transaction
                pass
            else:
                last_txn = txn
                retval.append(txn)
        return retval

    @staticmethod
    def extract_sort_key(txn):
        if hasattr(txn, "tradeDate"):
            return txn.tradeDate
        elif hasattr(txn, "date"):
            return txn.date
        elif hasattr(txn, "settleDate"):
            return txn.settleDate
        return None

    def filter(self, txns, acctid):
        if len(txns) == 0:
            sorted_txns = txns
        else:
            sorted_txns = sorted(txns, key=OfxSynchronizer.extract_sort_key)
        retval = [txn for txn in sorted_txns if not (self.is_txn_synced(acctid, txn))]
        return self.filter_comment_txns(retval)

    def get_new_txns(self, acct, max_days=999999, resync=False):
        if resync or (max_days < 7):
            days = max_days
        else:
            days = 7
        last_txns_len = 0
        while True:
            logging.debug(
                "Downloading %d days of transactions for %s (max_days=%d)."
                % (days, acct.description, max_days)
            )
            raw = acct.download(days=days)

            if raw.read() == "Server error occured.  Received HttpStatusCode of 400":
                raise Exception("Error connecting to account %s" % (acct.description))
            raw.seek(0)
            ofx = None
            try:
                ofx = OfxParser.parse(raw)
            except OfxParserException as ex:
                if ex.message == "The ofx file is empty!":
                    return (ofx, [])
                else:
                    raise ex
            if ofx.signon is not None:
                if ofx.signon.severity == "ERROR":
                    raise Exception(
                        "Error returned from server for %s: %s"
                        % (acct.description, ofx.signon.message)
                    )
            if not (hasattr(ofx, "account")):
                # some banks return this for no txns
                if days >= max_days:
                    logging.debug("Hit max days.")
                    # return None to let the caller know that we don't
                    # even have account info
                    return (None, None)
                else:
                    days = days * 2
                    if days > max_days:
                        days = max_days
                    logging.debug("empty account: increasing days ago to %d." % (days))
                    last_txns_len = 0
            else:
                txns = ofx.account.statement.transactions
                new_txns = self.filter(txns, ofx.account.account_id)
                logging.debug("txns: %d" % (len(txns)))
                logging.debug("new txns: %d" % (len(new_txns)))
                if (len(txns) > 0) and (last_txns_len == len(txns)):
                    # not getting more txns than last time; we have
                    # reached the beginning
                    logging.debug("Not getting more txns than last time, done.")
                    return (ofx, new_txns)
                elif (len(txns) > len(new_txns)) or (days >= max_days):
                    # got more txns than were new or hit max_days, we've
                    # reached a stopping point
                    if days >= max_days:
                        logging.debug("Hit max days.")
                    else:
                        logging.debug("Got some stale txns.")
                    return (ofx, new_txns)
                else:
                    # all txns were new, increase how far back we go
                    days = days * 2
                    if days > max_days:
                        days = max_days
                    logging.debug("Increasing days ago to %d." % (days))
                    last_txns_len = len(txns)


class CsvSynchronizer(Synchronizer):
    def __init__(self, lgr, payee_format=None, date_format=None):
        super(CsvSynchronizer, self).__init__(lgr)
        self.payee_format = payee_format
        self.date_format = date_format

    def is_row_synced(self, converter, row):
        if self.lgr is None:
            # User called with --no-ledger
            # All transactions are considered "synced" in this case.
            return False
        else:
            return self.lgr.check_transaction_by_id("csvid", converter.get_csv_id(row))

    def parse_file(self, path, accountname=None, unknownaccount=None):
        with open(path) as f:
            has_bom = f.read(3) == codecs.BOM_UTF8
            if not (has_bom):
                f.seek(0)
            else:
                f.seek(3)
            dialect = csv.Sniffer().sniff(f.readline())
            if not (has_bom):
                f.seek(0)
            else:
                f.seek(3)
            dialect.skipinitialspace = True
            reader = csv.DictReader(f, dialect=dialect)
            converter = CsvConverter.make_converter(
                set(reader.fieldnames),
                dialect,
                ledger=self.lgr,
                name=accountname,
                unknownaccount=unknownaccount,
                payee_format=self.payee_format,
                date_format=self.date_format,
            )
            # Create a new reader in case the converter modified the dialect
            if not (has_bom):
                f.seek(0)
            else:
                f.seek(3)
            reader = csv.DictReader(f, dialect=dialect)
            return [
                converter.convert(row)
                for row in reader
                if not (self.is_row_synced(converter, row))
            ]



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.


import os
import os.path
import tempfile

import pytest

from ledgerautosync.ledgerwrap import Ledger, LedgerPython


@pytest.mark.lgr_file("checking.lgr")
def test_check_transaction(ledger):
    assert ledger.check_transaction_by_id("ofxid", "1101.1452687~7.0000486")


@pytest.mark.lgr_file("checking.lgr")
def test_nonexistent_transaction(ledger):
    assert not ledger.check_transaction_by_id("ofxid", "FOO")


@pytest.mark.lgr_file("checking.lgr")
def test_empty_transaction(ledger):
    assert ledger.check_transaction_by_id("ofxid", "empty")


@pytest.mark.lgr_file("checking.lgr")
def test_get_account_by_payee(ledger):
    account = ledger.get_account_by_payee(
        "AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )", exclude="Assets:Foo"
    )
    assert account == "Expenses:Bar"


@pytest.mark.lgr_file("checking-dynamic-account.lgr")
def test_get_ambiguous_account_by_payee(ledger):
    account = ledger.get_account_by_payee("Generic", exclude="Assets:Foo")
    # shoud use the latest
    assert account == "Expenses:Bar"


@pytest.mark.lgr_file("checking.lgr")
def test_ofx_payee_quoting(ledger):
    payees = [
        "PAYEE TEST/SLASH",
        "PAYEE TEST,COMMA",
        "PAYEE TEST:COLON",
        "PAYEE TEST*STAR",
        "PAYEE TEST#HASH",
        "PAYEE TEST.PERIOD",
    ]
    for payee in payees:
        assert (
            ledger.get_account_by_payee(payee, ["Assets:Foo"]) is not None
        ), "Did not find %s in %s" % (payee, ledger)


# TODO Broken on current hledger
@pytest.mark.lgr_file("checking.lgr")
@pytest.mark.ledger_impls([Ledger, LedgerPython])
def test_ofx_payee_quote_quote(ledger):
    payees = [
        'PAYEE TEST"QUOTE',
    ]
    for payee in payees:
        assert (
            ledger.get_account_by_payee(payee, ["Assets:Foo"]) is not None
        ), "Did not find %s in %s" % (payee, ledger)


@pytest.mark.lgr_file("checking.lgr")
def test_ofx_id_quoting(ledger):
    assert (
        ledger.check_transaction_by_id("ofxid", "1/2") is True
    ), "Did not find 1/2 in %s" % (ledger)


@pytest.mark.lgr_file("checking.lgr")
def test_load_payees(ledger):
    ledger.load_payees()
    assert ledger.payees["PAYEE TEST:COLON"] == ["Assets:Foo", "Income:Bar"]


@pytest.mark.lgr_file("empty.lgr")
def test_load_payees_with_empty_ledger(ledger):
    ledger.load_payees()


# class TestLedger(LedgerTest, TestCase):
#     def setUp(ledger):
#         self.empty_lgr = Ledger(os.path.join("fixtures", "empty.lgr"), no_pipe=True)
#         ledger = Ledger(self.ledger_path, no_pipe=True)
#         self.dynamic_lgr = Ledger(self.dynamic_ledger_path, no_pipe=True)


@pytest.mark.lgr_file("checking.lgr")
@pytest.mark.ledger_impls([Ledger])
def test_args_only(ledger):
    (f, tmprcpath) = tempfile.mkstemp(".ledgerrc")
    os.close(f)  # Who wants to deal with low-level file descriptors?
    # Create an init file that will narrow the test data to a period that
    # contains no trasnactions
    with open(tmprcpath, "w") as f:
        f.write("--period 2012")
    # If the command returns no trasnactions, as we would expect if we
    # parsed the init file, then this will throw an exception.
    next(ledger.run([""]))
    os.unlink(tmprcpath)



# ledger-autosync plugin for CSV files from N26, a Berlin-based online bank.

import datetime
from decimal import Decimal
import re

from ledgerautosync.converter import Amount, Converter, CsvConverter, Posting, Transaction

class N26Converter(CsvConverter):
    FIELDSET = set(["Date","Payee","Account number","Transaction type","Payment reference","Amount (EUR)",
                    "Amount (Foreign Currency)","Type Foreign Currency","Exchange Rate"])

    def __init__(self, *args, **kwargs):
        super(N26Converter, self).__init__(*args, **kwargs)

    def mk_currency(self, currency):
        if currency == "USD":
            currency = "$"
        elif currency == "GBP":
            currency = "£"
        elif currency == "EUR":
            currency = "€"
        return currency

    def convert(self, row):
        amt = Decimal(row["Amount (EUR)"])
        curr_foreign = self.mk_currency(row["Type Foreign Currency"] or "EUR")
        amt_foreign = Decimal(row["Amount (Foreign Currency)"]) if row["Amount (Foreign Currency)"] else amt
        if amt < 0:
            reverse   = False
            acct_from = self.name
            curr_from = self.mk_currency("EUR")
            amt_from  = Amount(amt, curr_from, reverse=reverse)
            acct_to   = "Expenses:Misc"
            curr_to   = curr_foreign
            amt_to    = Amount(amt_foreign, curr_to, reverse=not reverse)
        else:
            reverse   = True
            acct_from = "Assets:Other"
            curr_from = curr_foreign
            amt_from  = Amount(amt_foreign, curr_from, reverse=reverse)
            acct_to   = self.name
            curr_to   = self.mk_currency("EUR")
            amt_to    = Amount(amt, curr_to,   reverse=not reverse)

        payee = re.sub(r"[A-Za-z]+('[A-Za-z]+)?",lambda word: word.group(0).capitalize(),row["Payee"])
        meta = {"csvid": self.get_csv_id(row)}

        posting_from = Posting(acct_from, amt_from, metadata=meta if acct_from == self.name else {})
        posting_to   = Posting(acct_to,   amt_to,   metadata=meta if acct_to   == self.name else {})

        return Transaction(
            date=datetime.datetime.strptime(row["Date"], "%Y-%m-%d"),
            cleared=True,
            date_format="%Y-%m-%d",
            payee=payee,
            postings=[posting_to, posting_from]
        )



# Copyright (c) 2013-2021 Erik Hetzner
#
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

import os.path
import re
import tempfile
from io import StringIO
from unittest.mock import Mock, call, patch

import pytest
from ofxclient.config import OfxConfig

from ledgerautosync import LedgerAutosyncException
from ledgerautosync.cli import find_ledger_file, run


def test_run():
    config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
    acct = config.accounts()[0]
    acct.download = Mock(
        side_effect=lambda *args, **kwargs: open(
            os.path.join("fixtures", "checking.ofx"), "rb"
        )
    )
    config.accounts = Mock(return_value=[acct])
    run(["-l", os.path.join("fixtures", "empty.lgr")], config)
    acct.download.assert_has_calls([call(days=7), call(days=14)])
    assert config.accounts.call_count == 1


def test_run_csv_file():
    config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
    run(
        [
            "-a",
            "Paypal",
            "-l",
            os.path.join("fixtures", "empty.lgr"),
            os.path.join("fixtures", "paypal.csv"),
        ],
        config,
    )


def test_filter_account():
    config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
    foo = next(
        acct for acct in config.accounts() if acct.description == "Assets:Savings:Foo"
    )
    bar = next(
        acct for acct in config.accounts() if acct.description == "Assets:Checking:Bar"
    )
    foo.download = Mock(
        side_effect=lambda *args, **kwargs: open(
            os.path.join("fixtures", "checking.ofx"), "rb"
        )
    )
    bar.download = Mock()
    config.accounts = Mock(return_value=[foo, bar])
    run(
        [
            "-l",
            os.path.join("fixtures", "checking.lgr"),
            "-a",
            "Assets:Savings:Foo",
        ],
        config,
    )
    foo.download.assert_has_calls([call(days=7)])
    bar.download.assert_not_called()


def test_find_ledger_path():
    os.environ["LEDGER_FILE"] = "/tmp/foo"
    assert (
        find_ledger_file() == "/tmp/foo"
    ), "Should use LEDGER_FILE to find ledger path."

    (f, tmprcpath) = tempfile.mkstemp(".ledgerrc")
    os.close(f)  # Who wants to deal with low-level file descriptors?
    with open(tmprcpath, "w") as f:
        f.write("--bar foo\n")
        f.write("--file /tmp/bar\n")
        f.write("--foo bar\n")
    assert (
        find_ledger_file(tmprcpath) == "/tmp/foo"
    ), "Should prefer LEDGER_FILE to --file arg in ledgerrc"
    del os.environ["LEDGER_FILE"]
    assert find_ledger_file(tmprcpath) == "/tmp/bar", "Should parse ledgerrc"
    os.unlink(tmprcpath)


def test_no_ledger_arg():
    with pytest.raises(LedgerAutosyncException):
        config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
        run(["-l", os.path.join("fixtures", "checking.lgr"), "-L"], config)


def test_format_payee():
    with patch("sys.stdout", new_callable=StringIO) as mock_stdout:
        run(
            [
                os.path.join("fixtures", "paypal.csv"),
                "-a",
                "Assets:Foo",
                "--payee-format",
                "GROSS:{Gross}",
                "-L",
            ]
        )
        assert re.search(r"GROSS:-20\.00", mock_stdout.getvalue())


# def test_multi_account():
#     with patch('sys.stdout', new_callable=StringIO) as mock_stdout:
#         run([os.path.join('fixtures', 'multi_account.ofx'), '-a', 'Assets:Foo'])
#         self.assertRegexpMatches(mock_stdout.getvalue(), r"GROSS:-20\.00")


def test_no_ledger():
    config = OfxConfig(os.path.join("fixtures", "ofxclient.ini"))
    acct = config.accounts()[0]
    acct.download = Mock(
        side_effect=lambda *args, **kwargs: open(
            os.path.join("fixtures", "checking.ofx"), "rb"
        )
    )
    config.accounts = Mock(return_value=[acct])
    with patch("ledgerautosync.cli.find_ledger_file", return_value=None):
        with patch("sys.stderr", new_callable=StringIO) as mock_stdout:
            run([], config)
            assert (
                mock_stdout.getvalue()
                == "LEDGER_FILE environment variable not set, and no .ledgerrc file found, and -l argument was not supplied: running with deduplication disabled. All transactions will be printed!\n"
            )



#!/usr/bin/python
import re
import sys

first_line = False
ofxline = None

with open(sys.argv[1]) as f:
    for line in f.readlines():
        md = re.match(r"^(19|20)[0-9][0-9]", line)
        if md is not None:
            # Mark the next line as the first line in a txn
            first_line = True
        else:
            if first_line:
                first_line = False
                # Check if there is an ofxid on this line
                md = re.match(r"^\s+; ofxid:", line)
                if md is not None:
                    ofxline = line
                    continue
        # In every case except the one above where we call next, print the line
        sys.stdout.write(line)
        # We had a misplaced ofxid last, print it now
        if ofxline:
            sys.stdout.write(ofxline)
            ofxline = None
